--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: CSV
-- @author    Valentin Schmidt
-- @version   0.3
--****************************************************************************

property $

----------------------------------------
-- @constructor
----------------------------------------
on new (me)

  $.import("string")

  return me
end

----------------------------------------
-- Parses CSV string into list of lists
-- Notice: NOT supporting <delimiter> inside enclosed fields yet!
-- @param {string} str
-- @param {string} [delimiter=TAB]
-- @param {string} [EOL=CRLF]
-- @param {string} [enclosure]
-- @param {bool} [rowsAsPropLists]
-- @return {list}
----------------------------------------
on parse (me, csvStr, delimiter, EOL, enclosure, rowsAsPropLists)

  if voidP(delimiter) then delimiter=TAB
  if voidP(EOL) then EOL=$.CRLF
  if voidP(enclosure) then enclosure=""
  if voidP(rowsAsPropLists) then rowsAsPropLists=FALSE

  res = []

  -- extract lines
  lines = $.string.explode(EOL, csvStr)

  -- remove trailing empty lines
  repeat while lines.getLast()="" then
    lines.deleteAt(lines.count)
  end repeat

  lineCnt = lines.count
  if lineCnt=0 then return []

  -- parse first line into list
  firstRow = me._parseLine(lines[1], delimiter, enclosure)
  colCnt = firstRow.count

  if rowsAsPropLists=TRUE then -- propList
    headers = firstRow
    repeat with i = 2 to lineCnt
      row = me._parseLine(lines[i], delimiter, enclosure)
      props = [:]
      repeat with i = 1 to colCnt
        props[headers[i]] = row[i]
      end repeat
      res.add(props)
    end repeat
  else -- list
    res.add(firstRow)
    repeat with i = 2 to lineCnt
      row = me._parseLine(lines[i], delimiter, enclosure)
      res.add(row)
    end repeat
  end if

  return res
end

----------------------------------------
--
----------------------------------------
on streamParse (me, fn, delimiter, EOL, enclosure, encoding, rowCallbackHandler, rowCallbackTarget)
  if voidP(delimiter) then delimiter=TAB
  if voidP(EOL) then delimiter=$.LF
  if voidP(rowCallbackTarget) then rowCallbackTarget = _movie
  if voidP(encoding) then encoding="utf-8"

--  me.pCbHandler = rowCallbackHandler
--  me.pCbTarget

  fp = $.file.fopen(fn, "r")

  fileLen = $.file.fsize(fp)

  -- check for BOM 0xef, 0xbb, 0xbf
  tmp = $.file.freadbytes(fp, 3)
  if tmp[1] = 239 then
    fileLen = fileLen-3
  else
    $.file.fseek(fp, 0)
  end if

--  -- 512: 8532
--  -- 1024: 7587
--  -- 2048: 7115
--  -- 4096: 6843
--  -- 8192: 6732
--  -- 16384: 6670
--  -- 32768: 6683
  bufSize = 16384
  cnt = fileLen/bufSize

  rows = 0
  lineStr = ""

  --ms = the milliseconds

  if EOL=$.CRLF then
    EOL = 13
    skip = 2
  else
    EOL = chartonum(EOL)
    skip = 1
  end if

  repeat with i = 1 to cnt

    -- read bufSize bytes into byteArray
    ba = $.file.freadbytes(fp, bufSize)

    lastPos = 1
    first = 1
    repeat with j = 1 to bufSize
      if ba[j]=EOL then -- EOL found

--        rows=rows+1
--        if rows=10 then return

        -- test
        if not first then ba.position = ba.position+skip

        -- read the string
        put ba.readRawString((j-1) - (lastPos-1) - skip+first, encoding) after lineStr

        call(rowCallbackHandler, rowCallbackTarget, me._parseLine(lineStr, delimiter, enclosure))

        lineStr = ""
        lastPos = j
        first = 0
      end if
    end repeat
  end repeat

  --last chunk
  rest = fileLen - cnt*bufSize
  if rest then
    ba = $.file.freadbytes(fp, rest)

    lastPos = 1
    first = 1
    repeat with j = 1 to rest

      if ba[j]=EOL then
        --rows=rows+1

        if not first then ba.position = ba.position+skip

        -- read the string
        put ba.readRawString((j-1) - (lastPos-1) - skip+first, encoding) after lineStr

        call(rowCallbackHandler, rowCallbackTarget, me._parseLine(lineStr, delimiter, enclosure))

        lineStr = ""
        lastPos = j
        first = 0
      end if

    end repeat
  end if

  $.file.fclose(fp)
end

----------------------------------------
-- Creates CSV string from list of lists (or propLists)
-- @todo
-- @param {list} data
-- @param {string} [delimiter=TAB]
-- @param {string} [EOL=CRLF]
-- @param {string} [enclosure]
----------------------------------------
on create (me, data, delimiter, EOL, enclosure)

  -- TODO!

end

----------------------------------------
-- Parses a CSV line string
-- @return {list}
----------------------------------------
on _parseLine (me, lineStr, delimiter, enclosure)
  cells = $.string.explode(delimiter, lineStr)

  if not voidP(enclosure) then
    cnt = cells.count
    repeat with i = 1 to cnt
      str = cells[i]

      -- remove enclosures
      if str starts enclosure then
        delete char 1 of str
        delete the last char of str
      end if

      -- replace double with single enclosure
      str = $.string.replace(enclosure&enclosure, enclosure, str)

      cells[i] = str
    end repeat
  end if

  return cells
end
